package org.processmining.subtrace.algorithms;

import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XTimeBounds;
import org.deckfour.xes.info.impl.XLogInfoImpl;
import org.deckfour.xes.model.*;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.log.utils.XUtils;
import org.processmining.subtrace.parameters.SubTraceFilteringParameters;

import java.util.Date;
import java.util.HashMap;
import java.util.Map;

public class SubTraceAlgorithm {

    /**
     * The method that implements your algorithm.
     *
     * @param context    The context where to run in.
     * @param incoming   The event log
     * @param parameters The parameters to use.
     * @return Event log with subtraces.
     */
    public XLog apply(PluginContext context, XLog incoming, SubTraceFilteringParameters parameters) {
        /**
         * algorithm extract fragments based on filter conditions
         */
        long time = -System.currentTimeMillis();
        parameters.displayMessage("[SubTrace] Start");
        XLogInfo logInfo = XLogInfoImpl.create(incoming, XLogInfoImpl.STANDARD_CLASSIFIER);
        parameters.displayMessage("[SubTrace] Input Log = " + XConceptExtension.instance().extractName(incoming) +
                " traces: " + logInfo.getNumberOfTraces() +
                " events: " + logInfo.getNumberOfEvents());
        parameters.displayMessage("[SubTrace] Parameters = " + parameters.toString());

        XLog output = filterLog(incoming, parameters, context);

        logInfo = XLogInfoImpl.create(output, XLogInfoImpl.STANDARD_CLASSIFIER);
        time += System.currentTimeMillis();
        parameters.displayMessage("[SubTrace] Output log = " + XConceptExtension.instance().extractName(output) +
                " traces: " + logInfo.getNumberOfTraces() +
                " events: " + logInfo.getNumberOfEvents());

        parameters.displayMessage("[SubTrace] End (took " + time / 1000.0 + "  seconds).");

        return output;
    }

    /**
     * Extract fragments based on filter conditions
     *
     * @param incoming   log to process
     * @param parameters parameters
     * @return log with fragments
     */
    private static XLog filterLog(XLog incoming, SubTraceFilteringParameters parameters, PluginContext context) {
        XLog outgoing = (XLog) incoming.clone();
        XConceptExtension.instance().assignName(outgoing, parameters.getOutgoingLogName());

        boolean startCondition = parameters.getStartEvents().size() > 0;
        boolean endCondition = parameters.getEndEvents().size() > 0;
        boolean requiredCondition = parameters.getRequiredEvents().size() > 0;
        boolean edgeCondition = !parameters.getStartEdgeProcessing().equals(SubTraceFilteringParameters.EdgeProcessingType.KEEP) ||
                                !parameters.getEndEdgeProcessing().equals(SubTraceFilteringParameters.EdgeProcessingType.KEEP);

        if (startCondition || endCondition || requiredCondition || edgeCondition || parameters.isCalculateMetrics()) {

            // if we have context = draw progress bar
            if (context != null) {
                context.getProgress().setCaption("Processing log");
                context.getProgress().setMinimum(0);
                context.getProgress().setMaximum(incoming.size());
                context.getProgress().setValue(0);
                context.getProgress().setIndeterminate(false);
            }

            int traceIndex = 0;
            while (traceIndex < outgoing.size()) {
                XTrace xTrace = outgoing.get(traceIndex);

                if (parameters.isCalculateMetrics()) {
                    XUtils.putAttribute(xTrace,
                            XUtils.createAttribute("variant:startEvent",
                                    parameters.getClassifier().getClassIdentity(xTrace.get(0))));

                    XUtils.putAttribute(xTrace,
                            XUtils.createAttribute("variant:endEvent",
                                    parameters.getClassifier().getClassIdentity(xTrace.get(xTrace.size() - 1))));

                    XUtils.putAttribute(xTrace,
                            XUtils.createAttribute("variant:eventCount",xTrace.size()));

                    Date traceStart = XUtils.getTimestamp(xTrace.get(0));
                    Date traceEnd = XUtils.getTimestamp(xTrace.get(xTrace.size()-1));


                    Long seconds = 0L;
                    if (traceStart != null && traceEnd != null) {
                        seconds = (traceEnd.getTime() - traceStart.getTime()) / 1000;
                    }

                    XUtils.putAttribute(xTrace,
                            XUtils.createAttribute("variant:eventDuration",seconds));



                }

                boolean foundFirst = !startCondition;
                if (!foundFirst)
                    foundFirst = filterFindFirst(outgoing, xTrace, parameters);

                if (foundFirst) {
                    boolean foundLast = (!endCondition && !requiredCondition);
                    if (!foundLast)
                        foundLast = filterFindLast(outgoing, xTrace, parameters);

                    if (foundLast) {
                        filterProcessEdgeEvents(xTrace, parameters);

                        if (parameters.isCalculateMetrics()) {
                            calculateMetrics(outgoing, xTrace, parameters);
                        }

                        traceIndex++;
                    }
                }

                if (context != null) {
                    context.getProgress().inc();
                }

                if ( traceIndex % 10000 == 0 ) {
                    parameters.displayMessage("Extracted " + traceIndex + "traces");
                }
            }
        }
        return outgoing;
    }

    /**
     * Assemble variant from trace events
     * @param log - log to process
     * @param xTrace - trace to process
     * @param parameters - processing parameters
     */
    private static void calculateMetrics(XLog log, XTrace xTrace, SubTraceFilteringParameters parameters) {
        // traverse trace and calculate "variant"
        String variant = "Start - ";
        for (XEvent xEvent: xTrace) {
            variant += parameters.getClassifier().getClassIdentity(xEvent) + " - ";
        }
        variant += "End";

        XAttributeMap attributeMap = log.getAttributes();

        // Problems with List attribute implementation in MapDB, so let's try with discrete attribute and it's attributes
        if (!attributeMap.containsKey("variant:variants")) {
            XAttributeDiscrete variants = (XAttributeDiscrete)XUtils.createAttribute("variant:variants", 1);
            log.getAttributes().put("variant:variants", variants);
            log.getAttributes().get("variant:variants").getAttributes().put("0", XUtils.createAttribute("0", variant));

            XUtils.putAttribute(xTrace, XUtils.createAttribute("variant:variant", "0"));

        } else {
            String key = null;
            XAttributeMap variants = attributeMap.get("variant:variants").getAttributes();
            for (Map.Entry<String, XAttribute> attribute : variants.entrySet()) {
                if (attribute.getValue().toString().equals(variant)) {
                    key = attribute.getKey();
                    break;
                }
            }

            if (key != null) {
                XUtils.putAttribute(xTrace, XUtils.createAttribute("variant:variant", key));
            }
            else {
                Integer nextKey = variants.size();
                log.getAttributes().get("variant:variants").getAttributes().put(nextKey.toString(), XUtils.createAttribute(nextKey.toString(), variant));
                // FIXME MapDB makes attribute unmutable. Find way around
                //((XAttributeDiscrete)log.getAttributes().get("variant:variants")).setValue(nextKey + 1);
                XUtils.putAttribute(xTrace, XUtils.createAttribute("variant:variant", nextKey.toString()));

            }
        }

    }

    /**
     * Add/rename process start/end events
     */
    private static void filterProcessEdgeEvents(XTrace trace, SubTraceFilteringParameters parameters) {

        if (
                parameters.getStartEdgeProcessing().equals(SubTraceFilteringParameters.EdgeProcessingType.RENAME) &&
                        !parameters.getStartEventName().isEmpty()
                ) {
            XUtils.assignConceptName(trace.get(0), parameters.getStartEventName());
        }

        if (
                parameters.getStartEdgeProcessing().equals(SubTraceFilteringParameters.EdgeProcessingType.ADD) &&
                        !parameters.getStartEventName().isEmpty()
                ) {
            XEvent event = (XEvent) trace.get(0).clone();
            XUtils.assignConceptName(event, parameters.getStartEventName());
            trace.add(0, event);
        }

        if (
                parameters.getEndEdgeProcessing().equals(SubTraceFilteringParameters.EdgeProcessingType.RENAME) &&
                        !parameters.getEndEventName().isEmpty()
                ) {
            XUtils.assignConceptName(trace.get(trace.size() - 1), parameters.getEndEventName());
        }

        if (
                parameters.getEndEdgeProcessing().equals(SubTraceFilteringParameters.EdgeProcessingType.ADD) &&
                        !parameters.getEndEventName().isEmpty()
                ) {
            XEvent event = (XEvent) trace.get(trace.size() - 1).clone();
            XUtils.assignConceptName(event, parameters.getEndEventName());
            trace.add(event);
        }

    }

    /**
     * Run last event filter - delete all events before the first one
     *
     * @param outgoing   Log to filter
     * @param xTrace     Trace to filter
     * @param parameters Filter conditions
     * @return true if event was found and required event was present
     */
    private static boolean filterFindLast(XLog outgoing, XTrace xTrace, SubTraceFilteringParameters parameters) {
        boolean endCondition = parameters.getEndEvents().size() > 0;
        boolean foundLast = !endCondition;
        boolean foundRequired = !(parameters.getRequiredEvents().size() > 0);

        int currentIndex = 0;
        int lastIndex = 0;
        XEvent current;
        String name;
        Map<String, Integer> requiredCounts = new HashMap<>();
        while (currentIndex < xTrace.size()) {

            if (!foundRequired) {
                current = xTrace.get(currentIndex);
                name = parameters.getClassifier().getClassIdentity(current);
                if (parameters.getRequiredEvents().contains(name)) {
                    Integer count = requiredCounts.get(name);
                    count = (count == null) ? 1 : count + 1;
                    requiredCounts.put(name, count);

                    if (parameters.getRequiredEventsMatching().equals(SubTraceFilteringParameters.MatchingType.ANY)) {
                        if (parameters.getRequiredEventCount() <= count) {
                            foundRequired = true;
                        }
                    }
                    else if (parameters.getRequiredEventsMatching().equals(SubTraceFilteringParameters.MatchingType.NONE)) {
                        if (parameters.getRequiredEventCount() <= count) {
                            break; // end, failed
                        }
                    }
                }
            }

            if (endCondition) {
                if (parameters.getEndRange().equals(SubTraceFilteringParameters.RangeType.PREVIOUS)) {
                    if (currentIndex >= (xTrace.size() - 1)) {
                        break;
                    }
                    current = xTrace.get(currentIndex + 1);
                } else if(parameters.getEndRange().equals(SubTraceFilteringParameters.RangeType.NEXT)) {
                    if (currentIndex >= (xTrace.size() - 1)) {
                        break;
                    }
                    current = xTrace.get(currentIndex);
                }
                else {
                    current = xTrace.get(currentIndex);
                }
                name = parameters.getClassifier().getClassIdentity(current);
                if (parameters.getEndEvents().contains(name)) {
                    foundLast = true;
                    lastIndex = currentIndex;
                    if (parameters.getEndPositionType().equals(SubTraceFilteringParameters.PositionType.FIRST)) {
                        break;
                    }
                }
            }

            currentIndex++;
        }

        if (!foundRequired && parameters.getRequiredEventsMatching().equals(SubTraceFilteringParameters.MatchingType.ALL)) {
            foundRequired = true;
            for (Integer value: requiredCounts.values()) {
                foundRequired &= (parameters.getRequiredEventCount() <= value);
            }
        }

        if (!foundRequired && parameters.getRequiredEventsMatching().equals(SubTraceFilteringParameters.MatchingType.NONE)) {
            foundRequired = true;
            for (Integer value: requiredCounts.values()) {
                foundRequired &= (parameters.getRequiredEventCount() > value);
            }
        }

        if (
                !foundLast ||
                        !foundRequired ||
                        (parameters.getStartRange().equals(SubTraceFilteringParameters.RangeType.PREVIOUS) && xTrace.size() < 2)
                ) {
            outgoing.remove(xTrace);
        } else {
            if (parameters.getEndRange().equals(SubTraceFilteringParameters.RangeType.NEXT)) {
                lastIndex++;
            }

            // cut tail
            while (lastIndex < (xTrace.size() - 1)) {
                xTrace.remove(lastIndex + 1);
            }

        }

        return foundLast && foundRequired;
    }

    /**
     * Run start event filter - delete all events before the first one
     *
     * @param outgoing   Log to filter
     * @param xTrace     Trace to filter
     * @param parameters Filter conditions
     * @return true if event was found
     */
    private static boolean filterFindFirst(XLog outgoing, XTrace xTrace, SubTraceFilteringParameters parameters) {
        boolean foundFirst = false;

        int currentIndex = 0;
        int lastIndex = 0;
        XEvent current;
        String name;
        while (currentIndex < xTrace.size()) {
            if (parameters.getStartRange().equals(SubTraceFilteringParameters.RangeType.PREVIOUS)) {
                if (currentIndex >= (xTrace.size() - 1)) {
                    break;
                }
                current = xTrace.get(currentIndex + 1);
            } else {
                current = xTrace.get(currentIndex);
            }
            name = parameters.getClassifier().getClassIdentity(current);
            if (parameters.getStartEvents().contains(name)) {
                foundFirst = true;
                lastIndex = currentIndex;
                if (parameters.getStartPositionType().equals(SubTraceFilteringParameters.PositionType.FIRST)) {
                    break;
                }
            }

            if (!foundFirst) {
                xTrace.remove(currentIndex);
            } else {
                currentIndex++;
            }
        }

        if (
                !foundFirst ||
                        (parameters.getStartRange().equals(SubTraceFilteringParameters.RangeType.NEXT) && xTrace.size() < 2)
                ) {
            outgoing.remove(xTrace);
        } else {
            // cut to last found if looking for LAST
            if (parameters.getStartPositionType().equals(SubTraceFilteringParameters.PositionType.LAST)) {
                while (lastIndex > 0) {
                    xTrace.remove(0);
                    lastIndex--;
                }
            }
            // cut one more for NEXT range
            if (parameters.getStartRange().equals(SubTraceFilteringParameters.RangeType.NEXT)) {
                xTrace.remove(0);
            }
        }

        return foundFirst;
    }
}
