package org.processmining.subtrace.dialogs;

import com.fluxicon.slickerbox.factory.SlickerFactory;
import org.apache.commons.lang3.ArrayUtils;
import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.framework.util.ui.widgets.ProMTextField;
import org.processmining.subtrace.parameters.SubTraceFilteringParameters;

import javax.swing.*;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.text.NumberFormatter;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

/**
 * Generates dialog for Required Event selection step
 * Created by nsitnikov on 23/02/16.
 */
public class SubTraceRequiredEventStepDialog extends SubTraceDialog {
    private static final long serialVersionUID = -60187716353523368L;

    /**
     * The JPanel that allows the user to set (a subset of) the parameters.
     *
     * @param context UI context
     * @param input log to process
     * @param parameters parameters to update
     * @param logInfo computed log info
     */
    public SubTraceRequiredEventStepDialog(UIPluginContext context, XLog input, SubTraceFilteringParameters parameters, XLogInfo logInfo) {
        super(context, input, parameters, logInfo);
    }

    @Override
    protected JComponent generateList(final SubTraceFilteringParameters parameters) {
        JComponent jComponent = null;
        XEventClasses eventClasses = logInfo.getEventClasses(parameters.getClassifier());
        if (eventClasses != null) {
            List<XEventClass> sortedEventClasses = new ArrayList<>(eventClasses.getClasses());
            Collections.sort(sortedEventClasses, new EventClassComparator());

            // TODO look for more elegant way
            List<Integer> selected = new ArrayList<>();
            for (int i=0; i < sortedEventClasses.size(); i++) {
                if (parameters.getRequiredEvents().contains(sortedEventClasses.get(i).getId())) {
                    selected.add(i);
                }
            }

            final JList list = new JList<>(sortedEventClasses.toArray());
            list.clearSelection();
            if (selected.size() > 0 )
                list.setSelectedIndices(ArrayUtils.toPrimitive(selected.toArray(new Integer[selected.size()])));

            list.addListSelectionListener(new ListSelectionListener() {
                public void valueChanged(ListSelectionEvent e) {
                    List<Object> selectedObjects = list.getSelectedValuesList();
                    List<String> requiredEvents = new ArrayList<>();
                    for( Object item: selectedObjects) {
                        requiredEvents.add(((XEventClass)item).getId());
                    }
                    parameters.setRequiredEvents(requiredEvents);
                }
            });
            jComponent = configureList(list,
                    "Required events",
                    "Selected event(s) must be present in filtered fragment. Empty list (alt-click to remove selection) will disable this filter",
                    true);
        };

        return jComponent;
    }

    @Override
    protected JComponent generateRange(final SubTraceFilteringParameters parameters) {
        final ActionListener rangeListener = new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                if ("any".equals(e.getActionCommand())) {
                    parameters.setRequiredEventsMatching(SubTraceFilteringParameters.MatchingType.ANY);
                } else if ("all".equals(e.getActionCommand())) {
                    parameters.setRequiredEventsMatching(SubTraceFilteringParameters.MatchingType.ALL);
                } else if ("none".equals(e.getActionCommand())) {
                    parameters.setRequiredEventsMatching(SubTraceFilteringParameters.MatchingType.NONE);
                }
            }
        };
        //Create the radio buttons.
        List<JRadioButton> buttons = new ArrayList<>();
        JRadioButton button = SlickerFactory.instance().createRadioButton("Any event");
        button.setActionCommand("any");
        button.addActionListener(rangeListener);
        button.setSelected(parameters.getRequiredEventsMatching().equals(SubTraceFilteringParameters.MatchingType.ANY));
        buttons.add(button);

        button = SlickerFactory.instance().createRadioButton("All events");
        button.setActionCommand("all");
        button.addActionListener(rangeListener);
        button.setSelected(parameters.getRequiredEventsMatching().equals(SubTraceFilteringParameters.MatchingType.ALL));
        buttons.add(button);

        button = SlickerFactory.instance().createRadioButton("None of events");
        button.setActionCommand("none");
        button.addActionListener(rangeListener);
        button.setSelected(parameters.getRequiredEventsMatching().equals(SubTraceFilteringParameters.MatchingType.NONE));
        buttons.add(button);

        return configurePanel(buttons, "Choose matching rule:");

    }

    @Override
    protected JComponent generatePosition(final SubTraceFilteringParameters parameters) {

        NumberFormat numberFormat = NumberFormat.getIntegerInstance();
        numberFormat.setGroupingUsed(false);
        numberFormat.setMaximumIntegerDigits(3);

        final JFormattedTextField requiredCount =  new JFormattedTextField(numberFormat);
        requiredCount.setValue(parameters.getRequiredEventCount());
        requiredCount.setPreferredSize(new Dimension(40, 20));
        requiredCount.setMaximumSize(new Dimension(40, 20));
        requiredCount.setHorizontalAlignment(SwingConstants.RIGHT);

        ActionListener textListener = new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                Integer value = (Integer) requiredCount.getValue();
                if (value == null || value < 1) value = 1;
                parameters.setRequiredEventCount(value);
            }
        };

        requiredCount.addActionListener(textListener);

        return configureTextField(requiredCount, "Minimum event count per trace:");
    }

    @Override
    protected JComponent generateExtraEvent(final SubTraceFilteringParameters parameters) {
        final JCheckBox metrics = SlickerFactory.instance().createCheckBox("Assign variant id to traces", parameters.isCalculateMetrics());
        metrics.setFont(metrics.getFont().deriveFont(11f));

        ActionListener checkboxListener = new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                parameters.setCalculateMetrics(metrics.isSelected());
            }
        };

        metrics.addActionListener(checkboxListener);
        return metrics;
    }

    @Override
    protected JComponent generateExtraName(SubTraceFilteringParameters parameters) {
        return null;
    }

    @Override
    protected JComponent generateOutputName(SubTraceFilteringParameters parameters) {
        return null;
    }
}
