package org.processmining.subtrace.plugins;

import java.lang.reflect.Constructor;
import java.util.*;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.impl.XLogInfoImpl;
import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.*;
import org.processmining.log.utils.XLogBuilder;
import org.processmining.subtrace.algorithms.SubTraceAlgorithm;
import org.processmining.subtrace.connections.SubTraceConnection;
import org.processmining.subtrace.dialogs.*;
import org.processmining.subtrace.help.SubTraceHelp;
import org.processmining.subtrace.parameters.SubTraceFilteringParameters;

import javax.swing.*;

@Plugin(name = "Extract trace fragments", returnLabels = { "Extracted fragments", "Processing parameters" },
		returnTypes = {XLog.class, SubTraceFilteringParameters.class },
		parameterLabels = {"Log", "Processing parameters" }, level= PluginLevel.NightlyBuild, quality = PluginQuality.Good,
		categories = { PluginCategory.Filtering },
		userAccessible = true,
		help= SubTraceHelp.TEXT
)
public class SubTracePlugin extends SubTraceAlgorithm {

    //TODO add function to use any attribute for filtering (different for each screen)
    //TODO check if list scrollbar works
    //TODO make entry fields nicer
    //TODO find way to autotest dialogs
    //TODO add log metadata - list of variants to log attributes

    /**
     * The plug-in variant that runs in a UI context and allows to modify preset parameters.
     *
     * @param context The context to run in.
     * @param incoming The log to process.
     * @return The output.
     */
    @UITopiaVariant(affiliation = "RAMAX",
            author = "Nikolai Sitnikov",
            email = "nikolai.sitnikov@ramax.ru",
            pack = "SubTrace"
    )
    @PluginVariant(variantLabel = "Extract trace fragments", requiredParameterLabels = { 0, 1 },
            help = SubTraceHelp.TEXT_PARAMETERS)
    public Object[] runUIParameters(UIPluginContext context, XLog incoming, SubTraceFilteringParameters parameters) {
        // cache information for wizard
        XLogInfo logInfo = XLogInfoImpl.create(incoming, parameters.getClassifier());

        // Get list of dialog classes
        List<Class> steps = new ArrayList<>();
        steps.add(SubTraceEventClassifierStepDialog.class);
        steps.add(SubTraceStartEventStepDialog.class);
        steps.add(SubTraceEndEventStepDialog.class);
        steps.add(SubTraceRequiredEventStepDialog.class);

        // ListIterator iterator has very strange behavior, so I use index moving
        int currentIndex = 1;

        // Show the dialog. User can now change the parameters.
        while (true) {
            try {
                // Show the current step.
                Constructor<?> screen = steps.get(currentIndex - 1).getConstructor(UIPluginContext.class, XLog.class, SubTraceFilteringParameters.class, XLogInfo.class);
                InteractionResult result = context.showWizard(
                        "Configure trace fragments extraction. Step " + currentIndex + " of " + steps.size(),
                        currentIndex == 1,
                        currentIndex == steps.size(),
                        (JPanel)screen.newInstance(new Object[] {context, incoming, parameters, logInfo}));
                switch (result) {
                    case NEXT:
                        currentIndex++;
                        break;
                    case PREV:
                        currentIndex--;
                        break;
                    case FINISHED:
                        return new Object[]{runConnections(context, incoming, parameters), parameters};
                    default:
                        // Dialog got canceled.
                        return null;
                }
            }
            catch (Exception e) {
                e.printStackTrace();
            }
        }
    }


	/**
	 * The plug-in variant that runs in a UI context and uses a dialog to get the parameters.
	 * 
	 * @param context The context to run in.
	 * @param incoming The log to process.
	 * @return The output.
	 */
    @UITopiaVariant(affiliation = "RAMAX",
            author = "Nikolai Sitnikov",
            email = "nikolai.sitnikov@ramax.ru",
            pack = "SubTrace"
    )
	@PluginVariant(variantLabel = "Extract trace fragments, preset parameters", requiredParameterLabels = { 0 })
	public Object[] runUI(UIPluginContext context, XLog incoming) {
		// Get the default parameters.
	    SubTraceFilteringParameters parameters = new SubTraceFilteringParameters();
        parameters.setOutgoingLogName(XConceptExtension.instance().extractName(incoming) + " (fragments)");
        return runUIParameters(context, incoming, parameters);
	}

    /**
     * The plug-in variant that runs in a UI context and uses a dialog to get the parameters.
     *
     * @param context The context to run in.
     * @param incoming The log to process.
     * @return The output.
     */
    @UITopiaVariant(affiliation = "RAMAX",
            author = "Nikolai Sitnikov",
            email = "nikolai.sitnikov@ramax.ru",
            pack = "SubTrace"
    )
    @PluginVariant(variantLabel = "Extract trace fragments, preset parameters", requiredParameterLabels = { 0, 1 },
            help = SubTraceHelp.TEXT_BATCH
    )
    public Object[] runBatch(PluginContext context, XLog incoming, SubTraceFilteringParameters parameters) {
        // Get the default parameters.
        return  new Object[] { runConnections(context, incoming, parameters), parameters };
    }

	/**
	 * The plug-in variant that allows one to test the dialog to get the parameters.
	 * 
	 * @param context The context to run in.
	 * @return The output.
	 */
    @UITopiaVariant(affiliation = "RAMAX",
            author = "Nikolai Sitnikov",
            email = "nikolai.sitnikov@ramax.ru",
            pack = "SubTrace"
    )

	/**
	 * Apply the algorithm depending on whether a connection already exists.
	 * 
	 * @param context The context to run in.
	 * @param incoming Event log to process.
	 * @return XLog with subtraces.
	 */
	private XLog runConnections(PluginContext context, XLog incoming, SubTraceFilteringParameters parameters) {
		if (parameters.isTryConnections()) {
			// Try to found a connection that matches the inputs and the parameters.
			Collection<SubTraceConnection> connections;
			try {
				connections = context.getConnectionManager().getConnections(
						SubTraceConnection.class, context, incoming);
				for (SubTraceConnection connection : connections) {
					if (connection.getObjectWithRole(SubTraceConnection.FIRSTINPUT)
							.equals(incoming) && connection.getParameters().equals(parameters)) {
						// Found a match. Return the associated output as result of the algorithm.
						return connection
								.getObjectWithRole(SubTraceConnection.FIRSTOUTPUT);
					}
				}
			} catch (ConnectionCannotBeObtained e) {
			}
		}
		// No connection found. Apply the algorithm to compute a fresh output result.
		XLog output = apply(context, incoming, parameters);
		if (parameters.isTryConnections()) {
			// Store a connection containing the inputs, output, and parameters.
			context.getConnectionManager().addConnection(
					new SubTraceConnection(incoming, output, parameters));
		}
		// Return the output.
		return output;
	}
}
